/* Hexwars - A Hexagonal Turn Based Strategy Game
 * Copyright (C) 2009-2011  Patrik Schönfeldt
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "Engine.hpp"

Engine::Engine(Map* omap, UnitContainer* ounits, UI_Engine* uie, PLAYER oplyr){
    plyr = oplyr;
    map = omap;
    units = ounits;
    display = uie;
    ui = uie;
    tilesize = display->getTileSize();
    screensize = display->getSize();
    mapsize = map->getSize();

    if (screensize.x > (int) ((mapsize.x+0.25)*tilesize.x*0.75)){
        viewport.x = (-1)*(int) (screensize.x - (mapsize.x+0.25)*tilesize.x*0.75)/2;
        screensize.x = (int) ((mapsize.x+0.25)*tilesize.x*0.75);
        scroll_x = false;
    } else {
        scroll_x = true;
    }
    if (screensize.y > (int)((mapsize.y + 0.5) * tilesize.y)){
        viewport.y = (-1)*(int)(screensize.y - (mapsize.y + 0.5) * tilesize.y)/2;
        screensize.y = (int)((mapsize.y + 0.5) * tilesize.y);
        scroll_y = false;
    } else {
        scroll_y = true;
    }
    maxviewport.x = ((mapsize.x + 0.25)*tilesize.x*0.75) - screensize.x;
    maxviewport.y = ((mapsize.y + 0.5) * tilesize.y) - screensize.y;
}

Engine::~Engine(){

}

void Engine::drawLine(std::vector<Coordinates> line, int col){
    // transform pos in tiles to pos on screen
    for (unsigned int i=0; i<line.size(); i++){
        line[i] = displayPos(line[i]) + tilesize/2;
    }

    // draw a line for the way
    display->drawLine(line, col);
}

Coordinates Engine::displayPos(Coordinates pos){
    pos.y = pos.y * tilesize.y - viewport.y;
    if (pos.x % 2 != 0){
        pos.y += tilesize.y/2;
    }
    pos.x = pos.x * tilesize.x*3/4 - viewport.x;
    return pos;
}

void Engine::switchPlayer(PLAYER number){
    plyr = number;
    refresh();
}

void Engine::drawTerrain(TERRAINTYPE type, Coordinates pos){
    pos = displayPos(pos);
    display->drawTerrain(type, pos);
}

void Engine::drawSelection(Coordinates pos){
    pos = displayPos(pos);
    display->drawSelection(pos);
}

void Engine::drawUnit(Unit* unit){
    UNITTYPE type = unit->getType();
    p1 = displayPos(unit->getPos());
    p1.x += tilesize.x/2;
    display->drawUnit(type, unit->getOwner(), p1);

    // draw bar for old healthpoints
    if (unit->getOldHp() > 0){
        p1.x += tilesize.x/5+1;
        p1.y += tilesize.y*3/4;
        p2 = p1;
        short length = unit->getOldHp();
        while (length > tilesize.y/4*3){
            p2.y -= tilesize.y/4*3;
            display->drawLine(p1,p2, 0xff7777);
            p1.x += 2;
            p2 = p1;
            length -= tilesize.y/4*3;
        }
        p2.y -= length;
        display->drawLine(p1,p2, 0xff7777);
    }

    // draw bar for healthpoints
    p1 = displayPos(unit->getPos());
    p1.x += tilesize.x/2;
    if (unit->getHp() > 0){
        p1.x += tilesize.x/5;
        p1.y += tilesize.y*3/4;
        p2 = p1;
        short length = unit->getHp();
        while (length > tilesize.y/4*3){
            p2.y -= tilesize.y/4*3;
            display->drawLine(p1,p2, 0xff0000);
            p1.x += 2;
            p2 = p1;
            length -= tilesize.y/4*3;
        }
        p2.y -= length;
        display->drawLine(p1,p2, 0xff0000);
    }
}

void Engine::drawUnit(UNITTYPE type, PLAYER p, Coordinates pos){
    pos = displayPos(pos);
    display->drawUnit(type, p, pos);
}

void Engine::refresh(){
    selectedunit = units->getSelectedUnit();

    // locking the screen might enhance drawing performance
    display->lockScreen(true);

    // first of all clear the screen
    display->clearBlack();

     // mark all fields invisible
     for(p1.x =0; p1.x<mapsize.x; (p1.x)++){
          for(p1.y =0; p1.y<mapsize.y; (p1.y)++){
               visible[p1.x][p1.y] = 0;
          }
     }

    /* check if fields are visible
     * and mark them */
    if(plyr != PLAYER0){
        for(unsigned int u=1; u <= units->getLatestUuid(); u++){
            drawthis = units->getUnit(u);
            if (drawthis != NULL){
                if (drawthis->getOwner() == plyr && drawthis->getOldHp() > 0){
                    p1 = drawthis->getPos();
                    for(p2.x=0; p2.x<mapsize.x; (p2.x)++){
                        for(p2.y=0; p2.y<mapsize.y; (p2.y)++){
                            if( visible[p2.x][p2.y]
                                    < (int) drawthis->getSight()
                                    - (int) p1.distanceTo(p2)) {
                                visible[p2.x][p2.y]
                                        = (int) drawthis->getSight()
                                        - (int) p1.distanceTo(p2);
                            }
                        }
                    }
                }
            }
        }
    } else {
        for(p2.x=0; p2.x<mapsize.x; (p2.x)++){
            for(p2.y=0; p2.y<mapsize.y; (p2.y)++){
                visible[p2.x][p2.y] = 2;
            }
        }
    }


    /* draw all visible terrain
      */
    for( p1.y = (viewport.y/tilesize.y)-1
                ; (p1.y-1)*tilesize.y < screensize.y + viewport.y
                ; (p1.y)++) {
        for( p1.x = (viewport.x/tilesize.x)*2/3 -1
                    ; (p1.x*((int)(0.375* (float)tilesize.x))
                       < screensize.x + viewport.x)
                    ; (p1.x)++) {
            if (Coordinates(-1,-1) < p1 and p1 < mapsize){
                if (visible[p1.x][p1.y] > 0){
                    drawTerrain(map->getTerrain(p1), p1);
                    if(visible[p1.x][p1.y] == 1){
                        drawUnit(FOG, 0, p1);
                    }
                } else {
                    drawTerrain(BLACK, p1);
                }
            }
        }
    }

    // draw oldways of player's units
    for(unsigned int u=1; u <= units->getLatestUuid(); u++){
        drawthis = units->getUnit(u);
        if (drawthis != NULL && drawthis->getOwner() == plyr){
            way = drawthis->getOldWay();
            way.push_back(drawthis->getPos());
            drawLine(way, 0x7f7f7f);
        }
    }

    // draw ways of player's units
    for(unsigned int u=1; u <= units->getLatestUuid(); u++){
        drawthis = units->getUnit(u);
        if (drawthis != NULL && drawthis->getOwner() == plyr && drawthis->isAlive()){
            way = drawthis->getWay();
            way.insert(way.begin(), drawthis->getPos());
            drawLine(way, 0xffff00);
        }
    }

    /* draw the way of selected unit (again),
     * (now) in red color */
    if (selectedunit != NULL && selectedunit->getOwner() == plyr){
        way = selectedunit->getWay();
        way.insert(way.begin(), selectedunit->getPos());
        drawLine(way, 0xff0000);
    }

    // draw selection
    if (selectedunit != NULL && selectedunit->getOwner() == plyr){
        drawSelection(selectedunit->getPos());
    }

    // draw all visible units
    for(unsigned int u=1; u <= units->getLatestUuid(); u++){
        drawthis = units->getUnit(u);
        if (drawthis != NULL){
            p1 = drawthis->getPos();
            if( visible[p1.x][p1.y] > 1){
                drawUnit(drawthis);
            }
        }
    }

    // need to release screen again
    display->lockScreen(false);
}

bool Engine::scroll() {
    /* pointerpos may be (-1,-1) if pointer is outside
     * the window, we check for this first because
     * scrolling might look strange in that case */
    if (pointerpos == Coordinates(-1,-1)){
        return false;
    }
    bool scrolled = false;
    if (scroll_x){
        //left border
        if (pointerpos.x < 32){
            if (viewport.x >= 5){
                viewport.x-=5;
                scrolled = true;
            } else if (viewport.x > 0){
                viewport.x = 0;
                scrolled = true;
            }
        }
        // right border
        if (pointerpos.x > screensize.x-32 && viewport.x >= 0){
            if(viewport.x + screensize.x +5 <= maxviewport.x + screensize.x){
                viewport.x+=5;
                scrolled = true;
            } else if (viewport.x < maxviewport.x){
                viewport.x = maxviewport.x;
                scrolled = true;
            }
        }
    }
    if (scroll_y){
        // upper border
        if (pointerpos.y < 32){
            if(viewport.y >= 5){
                viewport.y-=5;
                scrolled = true;
            } else if (viewport.y > 0){
                viewport.y=0;
                scrolled = true;
            }
        }
        // lower border
        if (pointerpos.y > screensize.y-32 && viewport.y >=0){
            if (viewport.y + screensize.y +5 <= maxviewport.y + screensize.y){
                viewport.y+=5;
                scrolled = true;
            } else if (viewport.y < maxviewport.y){
                viewport.y = maxviewport.y;
                scrolled = true;
            }
        }
    }
    return scrolled;
}

bool Engine::select(){
        selectedfield.x = (pointerpos.x+viewport.x-tilesize.x/4)/(tilesize.x*3/4);
        if (selectedfield.x%2 == 1){
            selectedfield.y = (pointerpos.y+viewport.y-tilesize.y/2)/tilesize.y;
        }
        else{
            selectedfield.y = (pointerpos.y+viewport.y)/tilesize.y;
        }
        if(lastselectedfield != selectedfield){
            lastselectedfield = selectedfield;
            return true;
        }else{
            return false;
        }
}

void Engine::specificInput(INPUTTYPE parinput){
    pointerpos = display->getPointerPos();

    switch (parinput){
    case LMB:
        select();
        currinput = SELECT;
        return;
    case RMB:
    case LMBHOLD:
    case RMBHOLD:
        select();
        currinput = parinput;
        return;
    case KP1:
    case KP2:
    case KP3:
    case KP5:
    case KP7:
    case KP8:
    case KP9:
        giveOrder(parinput);
        refresh();
        currinput = ORDER;
        return;
    default:
        break;
    }
    /* when nothing happened yet,
     * try if scrolling is done
     */
    if (scroll()){
        refresh();
        currinput = SCROLL;
    }
    return;
}

Coordinates Engine::getSelectedField(){
    return selectedfield;
}

void Engine::center(Coordinates c){
    if (scroll_x){
        viewport.x = tilesize.x * c.x * 3/4 + tilesize.x/2 - screensize.x/2;
        if( viewport.x < 0){
            viewport.x = 0;
        }
        if( viewport.x > maxviewport.x){
            viewport.x = maxviewport.x;
        }
    }
    if (scroll_y){
        if (c.x%2 == 1){
            viewport.y = tilesize.y * c.y * 3/4 + tilesize.y*2 - screensize.y/2;
        } else {
            viewport.y = tilesize.y * c.y * 3/4 + tilesize.y*3/2 - screensize.y/2;
        }
        if( viewport.y < 0){
            viewport.y = 0;
        }
        if( viewport.y > maxviewport.y){
            viewport.y = maxviewport.y;
        }
    }
}

void Engine::giveOrder(INPUTTYPE kp){
    selectedunit = units->getSelectedUnit();
    if ((selectedunit->getWay()).size() == 0){
        movefield = selectedunit->getPos();
    }
    else{
        movefield = ((selectedunit->getWay()).back());
    }
    if (selectedunit != NULL){
        if (selectedunit->getOwner() == plyr){
            switch(kp){
            case KP1:
                if (movefield.x%2 == 1){
                    movefield += Coordinates(-1, 1);
                }
                else{
                    movefield += Coordinates(-1, 0);
                }
                break;
            case KP2:
                movefield += Coordinates( 0, 1);
                break;
            case KP3:
                if (movefield.x%2 == 1){
                    movefield += Coordinates( 1, 1);
                }
                else{
                    movefield += Coordinates( 1, 0);
                }
                break;
            case KP5:
                movefield = selectedunit->getPos();
                break;
            case KP7:
                if (movefield.x%2 == 1){
                    movefield += Coordinates(-1, 0);
                }
                else{
                    movefield += Coordinates(-1, -1);
                }
                break;
            case KP8:
                movefield += Coordinates( 0, -1);
                break;
            case KP9:
                if (movefield.x%2 == 1){
                    movefield += Coordinates( 1, 0);
                }
                else{
                    movefield += Coordinates( 1,-1);
                }
                break;
            }
            if(Coordinates(-1,-1) < movefield && movefield < mapsize){
                selectedunit->command(movefield);
                currinput = MOVE;
            }
        }
    }
}
