/* Hexwars - A Hexagonal Turn Based Strategy Game
 * Copyright (C) 2009-2011  Patrik Schönfeldt
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/** This class needs implementations as
 *  real units. Objects should not be of type
 *  Unit itself. Constructor sets type of
 *  bare units to NOUNIT and speed, sight to 0.
 */

#ifndef _UNIT_H
#define _UNIT_H

typedef unsigned int UNITTYPE;
enum Units{
        NOUNIT=0, WARRIOR=1, SCOUT=2, TERRAFORMER=3, MAGICIAN=4, FOG
};

#include <vector>
#include <stdlib.h>
#include <time.h>

#include "../definitions/Coordinates.hpp"
#include "../definitions/Alias.hpp"
#include "../definitions/Map.hpp"

class Unit{
public:
    Unit();

    Unit(const Unit& orig);

    /** Reveals whether this Unit is alive or not.
     * @return false if the Unit has 0 hp, true otherwise
     */
    bool isAlive();

    void newRoundPoke();

    /** \short moves unit if a step is inteded at the given time
     *
     * The Unit moves along the given way. The Unit's speed defines
     * the maximum number of steps in a global scope but localy every
     * unit is pinged with timecodes an then moves if the timecode
     * hits a time when the unit is intended to take a step.
     *
     * @param time This is a value between 0 and 100, describing the
     * progress in the process of collateral Unit movement
     */
    virtual void move(unsigned int time);

    /** Gives a new command (field to go to). If current position equals
     * the given target, all commands will be canceled.
     * @param target is a Coordinate on the Map
     */
    void command(Coordinates);

    /** This method returns the current
     * position.
     * @return the Coordinates describing the position of this
     * Unit on the Map
     */
    Coordinates getPos();

    /**
     * \returns unit's owner
     */
    PLAYER getOwner();

    /** This method returns the current hitpoints
     *
     * @return current hitpoint value
     */
    short getHp();
    short getOldHp();

    /**
     * set units HP to 0
     */
    void kill();

    /** This method changes the current hitpoints by val
     */
    void changeHp(short val);

    /** This method returns current orders.
     * @return a vector of waypoints
     */
    std::vector<Coordinates> getWay();

    std::vector<Coordinates> getOldWay();

    void delOldWay();

    /** This method returns the unit's type.
     * @return the type of the Unit
     */
    UNITTYPE getType();

    /**
     * Getter for the Unit's sight distance
     */
    unsigned char getSight();

    /** Getter for the Unit's uuid.
     * @return the Units uuid
     */
    unsigned int getUuid();

    /** Setter for the Unit's uuid (only sets the uuid if it hasn't
     * been set so far).
     * @param id a valid uuid
     */
    void setUuid(unsigned int id);

    /**
     * connect the unit to a map
     */
    void setMap(Map*);

    void sync(Unit* orig);

protected:
    /** The Unit's type. New types need
     * declaration in "enum Units" (at
     * ../definitions/Alias.h) and their
     * own class declaration (at ./)
     */
    UNITTYPE type;

    /** The Unit's hitpoints/healthpoints
     */
    short hp;
    short oldhp;

    /**
     * the 'world' the unit lives on
     */
    Map* map;

    /** The Unit's owner
     */
    PLAYER owner;

    /** The Unit's speed given in fields per round.
     */
    unsigned char speed;

    /** The Unit's sight given in fields.
     */
    unsigned char sight;

    /** The unit's current position on the Map.
     */
    Coordinates pos;

    /** The fields to go to, when being moved.
     * New orders will be pushed to the end.
     */
    std::vector<Coordinates> way;
    std::vector<Coordinates> oldway;

    /** This is a unique ID for the player's Unit.
     * '0' means that there is no uuid set for this unit.
     */
    unsigned int uuid;

    /** use this in your constructor for implementations
     *  of class Unit
     */
    void init(Coordinates opos, PLAYER owner);
};

#endif    /* _UNIT_H */
